import click

from clinica.pipelines.engine import clinica_pipeline


@clinica_pipeline
@click.command(
    "deeplearning-prepare-data",
    deprecated=True,
    context_settings={
        "allow_extra_args": True,
        "ignore_unknown_options": True,
    },
)
def cli(*args, **kwargs):
    """Prepare data generated by Clinica for PyTorch with tensor extraction."""
    click.echo(
        "Functionalities provided by this pipeline have been moved to the `clinicadl` package."
        "\n\n"
        "See: https://clinicadl.readthedocs.io/en/stable/Preprocessing/Extract/"
    )
